from idaes.gas_solid_contactors.properties.oxygen_iron_OC_oxidation. \
      gas_phase_thermo import (GasPhaseParameterBlock,
                               GasPhaseStateBlock)
from idaes.gas_solid_contactors.properties.oxygen_iron_OC_oxidation. \
    solid_phase_thermo import (SolidPhaseParameterBlock,
                               SolidPhaseStateBlock)
from idaes.gas_solid_contactors.properties.oxygen_iron_OC_oxidation. \
    hetero_reactions import (HeteroReactionParameterBlock,
                             ReactionBlock)
